/// \file FT245RServer.h
///
///  Virtual access to host filesystem via CP/Net-style I/O device.
///
///  \date Apr 22, 2016
///  \author Douglas Miller
///

#ifndef FT245RSERVER_H_
#define FT245RSERVER_H_

#include "NetworkServer.h"
#include "ServerDispatch.h"
#include "propertyutil.h"

#include <stdint.h>
#include <string>

class FT245RServer : public ServerDispatch {
public:
	FT245RServer(PropertyUtil::PropertyMapT &props, std::string prefix, int clientId);
	virtual ~FT245RServer();
	int mainThread(int rxfd, int txfd);

protected:
	void startServer(int rxfd, int txfd);
	void resetServer();
private:
	void recvByte(uint8_t val, int txfd);
	void sendResp(int txfd, uint8_t *msgbuf, int len);
	static void *protoThread(void *v);

	uint8_t clientId;
	uint8_t hostId;

	static const int ndosLen = sizeof(struct NetworkServer::ndos);
	uint8_t buffer[256 + ndosLen];
	struct NetworkServer::ndos *header;
	int bufIx;
	int msgLen;
	int respLen;
	int dash;
	int value;

	pthread_t thread;
	int _rxfd;
	int _txfd;
};

#endif	// FT245RSERVER_H_
